"""
Here we provide a :class:`~moptipy.api.space.Space` of bin packings.

The bin packing object is defined in module
:mod:`~moptipyapps.binpacking2d.packing`. Basically, it is a
two-dimensional numpy array holding, for each item, its ID, its bin ID, as
well as its location defined by four coordinates.

1. Dequan Liu and Hongfei Teng. An Improved BL-Algorithm for Genetic Algorithm
   of the Orthogonal Packing of Rectangles. European Journal of Operational
   Research. 112(2):413-420. January (1999).
   https://doi.org/10.1016/S0377-2217(97)00437-2.
   http://www.paper.edu.cn/scholar/showpdf/MUT2AN0IOTD0Mxxh.
"""
from collections import Counter
from math import factorial
from typing import Final

import numpy as np
from moptipy.api.space import Space
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from pycommons.types import check_int_range, type_error

from moptipyapps.binpacking2d.instance import (
    IDX_HEIGHT,
    IDX_REPETITION,
    IDX_WIDTH,
    Instance,
)
from moptipyapps.binpacking2d.packing import (
    IDX_BIN,
    IDX_BOTTOM_Y,
    IDX_ID,
    IDX_LEFT_X,
    IDX_RIGHT_X,
    IDX_TOP_Y,
    Packing,
)
from moptipyapps.shared import SCOPE_INSTANCE


class PackingSpace(Space):
    """An implementation of the `Space` API of for 2D bin packings charts."""

    def __init__(self, instance: Instance) -> None:
        """
        Create a 2D packing space.

        :param instance: the 2d bin packing instance

        >>> inst = Instance.from_resource("a01")
        >>> space = PackingSpace(inst)
        >>> space.instance is inst
        True
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: The instance to which the packings apply.
        self.instance: Final[Instance] = instance

    def copy(self, dest: Packing, source: Packing) -> None:
        """
        Copy one packing to another one.

        :param dest: the destination packing
        :param source: the source packing
        """
        np.copyto(dest, source)
        dest.n_bins = source.n_bins

    def create(self) -> Packing:
        """
        Create a packing object without assigning items to locations.

        :return: the (empty, uninitialized) packing object

        >>> inst = Instance.from_resource("a01")
        >>> space = PackingSpace(inst)
        >>> x = space.create()
        >>> x.shape
        (24, 6)
        >>> x.instance is inst
        True
        >>> type(x)
        <class 'moptipyapps.binpacking2d.packing.Packing'>
        """
        return Packing(self.instance)

    def to_str(self, x: Packing) -> str:
        """
        Convert a packing to a string.

        :param x: the packing
        :return: a string corresponding to the flattened packing array

        >>> inst = Instance.from_resource("a01")
        >>> space = PackingSpace(inst)
        >>> y = space.create()
        >>> xx = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
        ...                1, 2, 2, 2, 2, 2, 2])
        >>> import moptipyapps.binpacking2d.encodings.ibl_encoding_1 as e
        >>> e._decode(xx, y, inst, inst.bin_width, inst.bin_height)
        5
        >>> space.to_str(y)
        '1;1;0;0;463;386;1;1;463;0;926;386;1;1;926;0;1389;386;\
1;1;1389;0;1852;386;1;1;1852;0;2315;386;1;1;0;386;463;772;\
1;1;463;386;926;772;1;1;926;386;1389;772;1;1;1389;386;1852;772;\
1;1;1852;386;2315;772;1;1;0;772;463;1158;1;1;463;772;926;1158;\
1;1;926;772;1389;1158;1;1;1389;772;1852;1158;1;1;1852;772;2315;1158;\
1;2;0;0;463;386;1;2;463;0;926;386;1;2;926;0;1389;386;2;2;0;386;1680;806;\
2;3;0;0;1680;420;2;3;0;420;1680;840;2;4;0;0;1680;420;2;4;0;420;1680;840;\
2;5;0;0;1680;420'
        """
        return CSV_SEPARATOR.join([str(xx) for xx in np.nditer(x)])

    def is_equal(self, x1: Packing, x2: Packing) -> bool:
        """
        Check if two bin packings have the same contents.

        :param x1: the first packing
        :param x2: the second packing
        :return: `True` if both packings are for the same instance and have the
            same structure

        >>> inst = Instance.from_resource("a01")
        >>> space = PackingSpace(inst)
        >>> y1 = space.create()
        >>> y1.fill(0)
        >>> y2 = space.create()
        >>> y2.fill(0)
        >>> space.is_equal(y1, y2)
        True
        >>> y1 is y2
        False
        >>> y1[0, 0] = 1
        >>> space.is_equal(y1, y2)
        False
        """
        return (x1.instance is x2.instance) and np.array_equal(x1, x2)

    def from_str(self, text: str) -> Packing:
        """
        Convert a string to a packing.

        :param text: the string
        :return: the packing

        >>> inst = Instance.from_resource("a01")
        >>> space = PackingSpace(inst)
        >>> y1 = space.create()
        >>> xx = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
        ...                1, 2, 2, 2, 2, 2, 2])
        >>> import moptipyapps.binpacking2d.encodings.ibl_encoding_1 as e
        >>> y1.n_bins = e._decode(
        ...     xx, y1, inst, inst.bin_width, inst.bin_height)
        >>> y2 = space.from_str(space.to_str(y1))
        >>> space.is_equal(y1, y2)
        True
        >>> y1 is y2
        False
        """
        if not isinstance(text, str):
            raise type_error(text, "packing text", str)
        x: Final[Packing] = self.create()
        np.copyto(x, np.fromstring(text, dtype=x.dtype, sep=CSV_SEPARATOR)
                  .reshape(x.shape))
        x.n_bins = int(x[:, IDX_BIN].max())
        self.validate(x)
        return x

    def validate(self, x: Packing) -> None:
        """
        Check if a packing is valid and feasible.

        This method performs a comprehensive feasibility and sanity check of
        a packing. It ensures that the packing could be implemented in the
        real world exactly as it is given here and that all data are valid and
        that it matches to the bin packing instance. This includes:

        - checking that all data types, numpy `dtypes`, and matrix shapes are
          correct
        - checking that the packing belongs to the same instance as this space
        - checking that no objects in the same bin overlap
        - checking that all objects occur exactly as often as prescribed by
          the instance
        - checking that no object extends outside of its bin
        - checking that all objects have the same width/height as prescribed
          in the instance (with possible rotations)
        - check that bin ids are assigned starting at `1` and incrementing in
          steps of `1`

        :param x: the packing
        :raises TypeError: if any component of the packing is of the wrong
            type
        :raises ValueError: if the packing is not feasible
        """
        if not isinstance(x, Packing):
            raise type_error(x, "x", Packing)
        inst: Final[Instance] = self.instance
        if inst is not x.instance:
            raise ValueError(
                f"x.instance must be {inst} but is {x.instance}.")
        if inst.dtype is not x.dtype:
            raise ValueError(
                f"inst.dtype = {inst.dtype} but x.dtype={x.dtype}")
        needed_shape: Final[tuple[int, int]] = (inst.n_items, 6)
        if x.shape != needed_shape:
            raise ValueError(f"x.shape={x.shape}, but must be {needed_shape}.")

        bin_width: Final[int] = check_int_range(
            inst.bin_width, "bin_width", 1, 1_000_000_000)
        bin_height: Final[int] = check_int_range(
            inst.bin_height, "bin_height", 1, 1_000_000_000)

        bins: Final[set[int]] = set()
        items: Final[Counter[int]] = Counter()

        for i in range(inst.n_items):
            item_id: int = int(x[i, IDX_ID])
            if (item_id <= 0) or (item_id > inst.n_different_items):
                raise ValueError(
                    f"Encountered invalid id={item_id} for object at index "
                    f"{i}, must be in 1..{inst.n_different_items}.")

            bin_id: int = int(x[i, IDX_BIN])
            if (bin_id <= 0) or (bin_id > inst.n_items):
                raise ValueError(
                    f"Encountered invalid bin-id={bin_id} for object at index"
                    f" {i}, must be in 1..{inst.n_items}.")
            bins.add(bin_id)

            items[item_id] += 1
            x_left: int = int(x[i, IDX_LEFT_X])
            y_bottom: int = int(x[i, IDX_BOTTOM_Y])
            x_right: int = int(x[i, IDX_RIGHT_X])
            y_top: int = int(x[i, IDX_TOP_Y])

            if (x_left >= x_right) or (y_bottom >= y_top):
                raise ValueError(
                    f"Invalid item coordinates ({x_left}, {y_bottom}, "
                    f"{x_right}, {y_top}) for item of id {item_id} at index"
                    f" {i}.")
            if (x_left < 0) or (y_bottom < 0) or (x_right > bin_width) \
                    or (y_top > bin_height):
                raise ValueError(
                    f"Item coordinates ({x_left}, {y_bottom}, {x_right}, "
                    f"{y_top}) for item of id {item_id} at index {i} extend "
                    f"outside of the bin ({bin_width}, {bin_height}).")

            real_width: int = x_right - x_left
            real_height: int = y_top - y_bottom
            width: int = int(inst[item_id - 1, IDX_WIDTH])
            height: int = int(inst[item_id - 1, IDX_HEIGHT])

            if ((real_width != width) or (real_height != height)) \
                    and ((real_width != height) and (real_height != width)):
                raise ValueError(
                    f"Item coordinates ({x_left}, {y_bottom}, {x_right}, "
                    f"{y_top}) mean width={real_width} and height="
                    f"{real_height} for item of id {item_id} at index {i}, "
                    f"which should have width={width} and height={height}.")

            for j in range(inst.n_items):
                if (int(x[j, IDX_BIN]) != bin_id) or (i == j):
                    continue
                x_left_2: int = int(x[j, IDX_LEFT_X])
                y_bottom_2: int = int(x[j, IDX_BOTTOM_Y])
                x_right_2: int = int(x[j, IDX_RIGHT_X])
                y_top_2: int = int(x[j, IDX_TOP_Y])
                if (x_left_2 < x_right) and (x_right_2 > x_left) \
                        and (y_bottom_2 < y_top) and (y_top_2 > y_bottom):
                    raise ValueError(
                        f"Item {x[j, IDX_ID]} in bin {bin_id} and at index "
                        f"{j} is ({x_left_2}, {y_bottom_2}, {x_right_2}, "
                        f"{y_top_2}) and thus intersects with item {item_id} "
                        f"at index {i} which is in the same bin at ({x_left},"
                        f" {y_bottom}, {x_right}, {y_top}).")

        for item_id, count in items.items():
            should: int = int(inst[item_id - 1, IDX_REPETITION])
            if should != count:
                raise ValueError(
                    f"Item {item_id} should occur {should} times, but occurs"
                    f" {count} times instead.")

        max_bin: Final[int] = max(bins)
        min_bin: Final[int] = min(bins)
        bin_count: Final[int] = len(bins)

        if (min_bin != 1) or ((max_bin - min_bin + 1) != bin_count):
            raise ValueError(
                f"Inconsistent use of bins. Found bins {sorted(bins)}.")

        if not isinstance(x.n_bins, int):
            raise type_error(x.n_bins, "x.n_bins", int)
        if x.n_bins != bin_count:
            raise ValueError(f"x.n_bins={x.n_bins}, but must be {bin_count}.")

    def n_points(self) -> int:
        """
        Get the number of possible packings.

        If we indeed consider that any object could be at any place, then
        there would be an incomprehensibly large number of possible packings.
        Here, we consider the bottom-left condition and the idea of encoding
        packings as signed permutations, as in the Liu and Teng paper "An
        Improved BL-Algorithm for Genetic Algorithm of the Orthogonal Packing
        of Rectangles." In this case, if `n` items are to be packed, the
        number of possible packings won't exceed `2^n * n!`.

        :return: the approximate number of packings

        >>> inst = Instance.from_resource("a01")
        >>> inst.n_items
        24
        >>> space = PackingSpace(inst)
        >>> space.n_points()
        10409396852733332453861621760000
        >>> from math import factorial
        >>> (2 ** 24) * factorial(24)
        10409396852733332453861621760000
        """
        return (2 ** self.instance.n_items) * factorial(self.instance.n_items)

    def __str__(self) -> str:
        """
        Get the name of the packing space.

        :return: the name

        >>> print(PackingSpace(Instance.from_resource("a10")))
        pack_a10
        """
        return f"pack_{self.instance}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.instance.log_parameters_to(kv)



"""
An extended end result record to represent packings.

This class extends the information provided by
:mod:`~moptipy.evaluation.end_results`. It allows us to compare the results
of experiments over different objective functions. It also represents the
bounds for the number of bins and for the objective functions. It also
includes the problem-specific information of two-dimensional bin packing
instances.
"""
import argparse
from dataclasses import dataclass
from math import isfinite
from typing import Any, Callable, Final, Iterable, Mapping, cast

from moptipy.api.objective import Objective
from moptipy.evaluation.base import (
    EvaluationDataElement,
)
from moptipy.evaluation.end_results import CsvReader as ErCsvReader
from moptipy.evaluation.end_results import CsvWriter as ErCsvWriter
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_results import from_logs as er_from_logs
from moptipy.evaluation.log_parser import LogParser
from pycommons.ds.immutable_map import immutable_mapping
from pycommons.ds.sequences import reiterable
from pycommons.io.console import logger
from pycommons.io.csv import (
    SCOPE_SEPARATOR,
    csv_column,
    csv_read,
    csv_scope,
    csv_select_scope,
    csv_write,
)
from pycommons.io.path import Path, file_path, line_writer
from pycommons.strings.string_conv import (
    num_to_str,
    str_to_num,
)
from pycommons.types import check_int_range, type_error

from moptipyapps.binpacking2d.instance import (
    Instance,
    _lower_bound_damv,
)
from moptipyapps.binpacking2d.objectives.bin_count import (
    BIN_COUNT_NAME,
    BinCount,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_empty import (
    BinCountAndEmpty,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_last_empty import (
    BinCountAndLastEmpty,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_last_skyline import (
    BinCountAndLastSkyline,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_last_small import (
    BinCountAndLastSmall,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_lowest_skyline import (
    BinCountAndLowestSkyline,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_small import (
    BinCountAndSmall,
)
from moptipyapps.binpacking2d.packing import Packing
from moptipyapps.shared import (
    moptipyapps_argparser,
    motipyapps_footer_bottom_comments,
)

#: the number of items
KEY_N_ITEMS: Final[str] = "nItems"
#: the number of different items
KEY_N_DIFFERENT_ITEMS: Final[str] = "nDifferentItems"
#: the bin width
KEY_BIN_WIDTH: Final[str] = "binWidth"
#: the bin height
KEY_BIN_HEIGHT: Final[str] = "binHeight"


#: the default objective functions
DEFAULT_OBJECTIVES: Final[tuple[Callable[[Instance], Objective], ...]] = (
    BinCount, BinCountAndLastEmpty, BinCountAndLastSmall,
    BinCountAndLastSkyline, BinCountAndEmpty, BinCountAndSmall,
    BinCountAndLowestSkyline)


def __lb_geometric(inst: Instance) -> int:
    """
    Compute the geometric lower bound.

    :param inst: the instance
    :return: the lower bound
    """
    area: Final[int] = sum(int(row[0]) * int(row[1]) * int(row[2])
                           for row in inst)
    bin_size: Final[int] = inst.bin_width * inst.bin_height
    res: int = area // bin_size
    return (res + 1) if ((res * bin_size) != area) else res


#: the lower bound of an objective
_OBJECTIVE_LOWER: Final[str] = "lowerBound"
#: the upper bound of an objective
_OBJECTIVE_UPPER: Final[str] = "upperBound"
#: the start string for bin bounds
LOWER_BOUNDS_BIN_COUNT: Final[str] = csv_scope("bins", _OBJECTIVE_LOWER)
#: the default bounds
_DEFAULT_BOUNDS: Final[Mapping[str, Callable[[Instance], int]]] = \
    immutable_mapping({
        LOWER_BOUNDS_BIN_COUNT: lambda i: i.lower_bound_bins,
        csv_scope(LOWER_BOUNDS_BIN_COUNT, "geometric"): __lb_geometric,
        csv_scope(LOWER_BOUNDS_BIN_COUNT, "damv"): lambda i: _lower_bound_damv(
            i.bin_width, i.bin_height, i),
    })


@dataclass(frozen=True, init=False, order=False, eq=False)
class PackingResult(EvaluationDataElement):
    """
    An end result record of one run of one packing algorithm on one problem.

    This record provides the information of the outcome of one application of
    one algorithm to one problem instance in an immutable way.
    """

    #: the original end result record
    end_result: EndResult
    #: the number of items in the instance
    n_items: int
    #: the number of different items in the instance
    n_different_items: int
    #: the bin width
    bin_width: int
    #: the bin height
    bin_height: int
    #: the objective values evaluated after the optimization
    objectives: Mapping[str, int | float]
    #: the bounds for the objective values (append ".lowerBound" and
    #: ".upperBound" to all objective function names)
    objective_bounds: Mapping[str, int | float]
    #: the bounds for the minimum number of bins of the instance
    bin_bounds: Mapping[str, int]

    def __init__(self,
                 end_result: EndResult,
                 n_items: int,
                 n_different_items: int,
                 bin_width: int,
                 bin_height: int,
                 objectives: Mapping[str, int | float],
                 objective_bounds: Mapping[str, int | float],
                 bin_bounds: Mapping[str, int]):
        """
        Create a consistent instance of :class:`PackingResult`.

        :param end_result: the end result
        :param n_items: the number of items
        :param n_different_items: the number of different items
        :param bin_width: the bin width
        :param bin_height: the bin height
        :param objectives: the objective values computed after the
            optimization
        :param bin_bounds: the different bounds for the number of bins
        :param objective_bounds: the bounds for the objective functions
        :raises TypeError: if any parameter has a wrong type
        :raises ValueError: if the parameter values are inconsistent
        """
        super().__init__()
        if not isinstance(end_result, EndResult):
            raise type_error(end_result, "end_result", EndResult)
        if end_result.best_f != objectives[end_result.objective]:
            raise ValueError(
                f"end_result.best_f={end_result.best_f}, but objectives["
                f"{end_result.objective!r}]="
                f"{objectives[end_result.objective]}.")
        if not isinstance(objectives, Mapping):
            raise type_error(objectives, "objectives", Mapping)
        if not isinstance(objective_bounds, Mapping):
            raise type_error(objective_bounds, "objective_bounds", Mapping)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        if len(objective_bounds) != (2 * len(objectives)):
            raise ValueError(f"it is required that there is a lower and an "
                             f"upper bound for each of the {len(objectives)} "
                             f"functions, but we got {len(objective_bounds)} "
                             f"bounds, objectives={objectives}, "
                             f"objective_bounds={objective_bounds}.")

        for name, value in objectives.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of evaluation[{name!r}]={value!r}", str)
            if not isinstance(value, int | float):
                raise type_error(
                    value, f"value of evaluation[{name!r}]={value!r}",
                    (int, float))
            if not isfinite(value):
                raise ValueError(
                    f"non-finite value of evaluation[{name!r}]={value!r}")
            lll: str = csv_scope(name, _OBJECTIVE_LOWER)
            lower = objective_bounds[lll]
            if not isfinite(lower):
                raise ValueError(f"{lll}=={lower}.")
            uuu = csv_scope(name, _OBJECTIVE_UPPER)
            upper = objective_bounds[uuu]
            if not (lower <= value <= upper):
                raise ValueError(
                    f"it is required that {lll}<=f<={uuu}, but got "
                    f"{lower}, {value}, and {upper}.")

        bins: Final[int | None] = cast(int, objectives[BIN_COUNT_NAME]) \
            if BIN_COUNT_NAME in objectives else None
        for name, value in bin_bounds.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of bounds[{name!r}]={value!r}", str)
            check_int_range(value, f"bounds[{name!r}]", 1, 1_000_000_000)
            if (bins is not None) and (bins < value):
                raise ValueError(
                    f"number of bins={bins} is inconsistent with "
                    f"bound {name!r}={value}.")

        object.__setattr__(self, "end_result", end_result)
        object.__setattr__(self, "objectives", immutable_mapping(objectives))
        object.__setattr__(self, "objective_bounds",
                           immutable_mapping(objective_bounds))
        object.__setattr__(self, "bin_bounds", immutable_mapping(bin_bounds))
        object.__setattr__(self, "n_different_items", check_int_range(
            n_different_items, "n_different_items", 1, 1_000_000_000_000))
        object.__setattr__(self, "n_items", check_int_range(
            n_items, "n_items", n_different_items, 1_000_000_000_000))
        object.__setattr__(self, "bin_width", check_int_range(
            bin_width, "bin_width", 1, 1_000_000_000_000))
        object.__setattr__(self, "bin_height", check_int_range(
            bin_height, "bin_height", 1, 1_000_000_000_000))

    def _tuple(self) -> tuple[Any, ...]:
        """
        Create a tuple with all the data of this data class for comparison.

        :returns: a tuple with all the data of this class, where `None` values
            are masked out
        """
        # noinspection PyProtectedMember
        return self.end_result._tuple()


def from_packing_and_end_result(  # pylint: disable=W0102
        end_result: EndResult, packing: Packing,
        objectives: Iterable[Callable[[Instance], Objective]] =
        DEFAULT_OBJECTIVES,
        bin_bounds: Mapping[str, Callable[[Instance], int]] =
        _DEFAULT_BOUNDS,
        cache: Mapping[str, tuple[Mapping[str, int], tuple[
            Objective, ...], Mapping[str, int | float]]] | None =
        None) -> PackingResult:
    """
    Create a `PackingResult` from an `EndResult` and a `Packing`.

    :param end_result: the end results record
    :param packing: the packing
    :param bin_bounds: the bounds computing functions
    :param objectives: the objective function factories
    :param cache: a cache that can store stuff if this function is to be
        called repeatedly
    :return: the packing result
    """
    if not isinstance(end_result, EndResult):
        raise type_error(end_result, "end_result", EndResult)
    if not isinstance(packing, Packing):
        raise type_error(packing, "packing", Packing)
    if not isinstance(objectives, Iterable):
        raise type_error(objectives, "objectives", Iterable)
    if not isinstance(bin_bounds, Mapping):
        raise type_error(bin_bounds, "bin_bounds", Mapping)
    if (cache is not None) and (not isinstance(cache, dict)):
        raise type_error(cache, "cache", (None, dict))

    instance: Final[Instance] = packing.instance
    if instance.name != end_result.instance:
        raise ValueError(
            f"packing.instance.name={instance.name!r}, but "
            f"end_result.instance={end_result.instance!r}.")

    row: tuple[Mapping[str, int], tuple[Objective, ...],
               Mapping[str, int | float]] | None = None \
        if (cache is None) else cache.get(instance.name, None)
    if row is None:
        objfs = tuple(sorted((obj(instance) for obj in objectives),
                             key=str))
        obounds = {}
        for objf in objfs:
            obounds[csv_scope(str(objf), _OBJECTIVE_LOWER)] = \
                objf.lower_bound()
            obounds[csv_scope(str(objf), _OBJECTIVE_UPPER)] = \
                objf.upper_bound()
        row = ({key: bin_bounds[key](instance)
                for key in sorted(bin_bounds.keys())}, objfs,
               immutable_mapping(obounds))
    if cache is not None:
        cache[instance.name] = row

    objective_values: dict[str, int | float] = {}
    bin_count: int = -1
    bin_count_obj: str = ""
    for objf in row[1]:
        z: int | float = objf.evaluate(packing)
        objfn: str = str(objf)
        objective_values[objfn] = z
        if not isinstance(z, int):
            continue
        if not isinstance(objf, BinCount):
            continue
        bc: int = objf.to_bin_count(z)
        if bin_count == -1:
            bin_count = bc
            bin_count_obj = objfn
        elif bin_count != bc:
            raise ValueError(
                f"found bin count disagreement: {bin_count} of "
                f"{bin_count_obj!r} != {bc} of {objf!r}")

    return PackingResult(
        end_result=end_result,
        n_items=instance.n_items,
        n_different_items=instance.n_different_items,
        bin_width=instance.bin_width, bin_height=instance.bin_height,
        objectives=objective_values,
        objective_bounds=row[2],
        bin_bounds=row[0])


def from_single_log(  # pylint: disable=W0102
        file: str,
        objectives: Iterable[Callable[[Instance], Objective]] =
        DEFAULT_OBJECTIVES,
        bin_bounds: Mapping[str, Callable[[Instance], int]] =
        _DEFAULT_BOUNDS,
        cache: Mapping[str, tuple[Mapping[str, int], tuple[
            Objective, ...], Mapping[str, int | float]]] | None =
        None) -> PackingResult:
    """
    Create a `PackingResult` from a file.

    :param file: the file path
    :param objectives: the objective function factories
    :param bin_bounds: the bounds computing functions
    :param cache: a cache that can store stuff if this function is to be
        called repeatedly
    :return: the packing result
    """
    the_file_path = file_path(file)
    end_results: Final[list[EndResult]] = []
    er_from_logs(the_file_path, end_results.append)
    if len(end_results) != 1:
        raise ValueError(
            f"needs one end result record in file {the_file_path!r}, "
            f"but got {end_results}.")

    packing = Packing.from_log(the_file_path)
    if not isinstance(packing, Packing):
        raise type_error(packing, f"packing from {file!r}", Packing)
    return from_packing_and_end_result(
        end_result=end_results[0], packing=packing,
        objectives=objectives, bin_bounds=bin_bounds, cache=cache)


def from_logs(  # pylint: disable=W0102
        directory: str,
        consumer: Callable[[PackingResult], None],
        objectives: Iterable[Callable[[Instance], Objective]] =
        DEFAULT_OBJECTIVES,
        bin_bounds: Mapping[str, Callable[[Instance], int]]
        = _DEFAULT_BOUNDS) -> None:
    """
    Parse a directory recursively to get all packing results.

    :param directory: the directory to parse
    :param consumer: the consumer for receiving the results
    :param objectives: the objective function factories
    :param bin_bounds: the bin bounds calculators
    """
    __LogParser(consumer, objectives, bin_bounds).parse_dir(directory)


def to_csv(results: Iterable[PackingResult], file: str) -> Path:
    """
    Write a sequence of packing results to a file in CSV format.

    :param results: the end results
    :param file: the path
    :return: the path of the file that was written
    """
    path: Final[Path] = Path(file)
    logger(f"Writing packing results to CSV file {path!r}.")
    path.ensure_parent_dir_exists()
    with path.open_for_write() as wt:
        consumer: Final[Callable[[str], None]] = line_writer(wt)
        for p in csv_write(
                data=sorted(results),
                setup=CsvWriter().setup,
                column_titles=CsvWriter.get_column_titles,
                get_row=CsvWriter.get_row,
                header_comments=CsvWriter.get_header_comments,
                footer_comments=CsvWriter.get_footer_comments,
                footer_bottom_comments=CsvWriter.get_footer_bottom_comments):
            consumer(p)
    logger(f"Done writing packing results to CSV file {path!r}.")
    return path


def from_csv(file: str) -> Iterable[PackingResult]:
    """
    Load the packing results from a CSV file.

    :param file: the file to read from
    :returns: the iterable with the packing result
    """
    path: Final[Path] = file_path(file)
    logger(f"Now reading CSV file {path!r}.")
    with path.open_for_read() as rd:
        yield from csv_read(
            rows=rd, setup=CsvReader, parse_row=CsvReader.parse_row)
    logger(f"Done reading CSV file {path!r}.")


class CsvWriter:
    """A class for CSV writing of :class:`PackingResult`."""

    def __init__(self, scope: str | None = None) -> None:
        """
        Initialize the csv writer.

        :param scope: the prefix to be pre-pended to all columns
        """
        #: an optional scope
        self.scope: Final[str | None] = (
            str.strip(scope)) if scope is not None else None

        #: has this writer been set up?
        self.__setup: bool = False
        #: the end result writer
        self.__er: Final[ErCsvWriter] = ErCsvWriter(scope)
        #: the bin bounds
        self.__bin_bounds: list[str] | None = None
        #: the objectives
        self.__objectives: list[str] | None = None

    def setup(self, data: Iterable[PackingResult]) -> "CsvWriter":
        """
        Set up this csv writer based on existing data.

        :param data: the data to setup with
        :returns: this writer
        """
        if self.__setup:
            raise ValueError("CSV writer has already been set up.")
        self.__setup = True

        data = reiterable(data)
        self.__er.setup(pr.end_result for pr in data)

        bin_bounds_set: Final[set[str]] = set()
        objectives_set: Final[set[str]] = set()
        for pr in data:
            bin_bounds_set.update(pr.bin_bounds.keys())
            objectives_set.update(pr.objectives.keys())
        if set.__len__(bin_bounds_set) > 0:
            self.__bin_bounds = sorted(bin_bounds_set)
        if set.__len__(objectives_set) > 0:
            self.__objectives = sorted(objectives_set)

        return self

    def get_column_titles(self) -> Iterable[str]:
        """
        Get the column titles.

        :returns: the column titles
        """
        p: Final[str] = self.scope
        yield from self.__er.get_column_titles()

        yield csv_scope(p, KEY_BIN_HEIGHT)
        yield csv_scope(p, KEY_BIN_WIDTH)
        yield csv_scope(p, KEY_N_ITEMS)
        yield csv_scope(p, KEY_N_DIFFERENT_ITEMS)
        if self.__bin_bounds:
            for b in self.__bin_bounds:
                yield csv_scope(p, b)
        if self.__objectives:
            for o in self.__objectives:
                oo: str = csv_scope(p, o)
                yield csv_scope(oo, _OBJECTIVE_LOWER)
                yield oo
                yield csv_scope(oo, _OBJECTIVE_UPPER)

    def get_row(self, data: PackingResult) -> Iterable[str]:
        """
        Render a single packing result record to a CSV row.

        :param data: the end result record
        :returns: the iterable with the row data
        """
        yield from self.__er.get_row(data.end_result)
        yield repr(data.bin_height)
        yield repr(data.bin_width)
        yield repr(data.n_items)
        yield repr(data.n_different_items)
        if self.__bin_bounds:
            for bb in self.__bin_bounds:
                yield (repr(data.bin_bounds[bb])
                       if bb in data.bin_bounds else "")
        if self.__objectives:
            for ob in self.__objectives:
                ox = csv_scope(ob, _OBJECTIVE_LOWER)
                yield (num_to_str(data.objective_bounds[ox])
                       if ox in data.objective_bounds else "")
                yield (num_to_str(data.objectives[ob])
                       if ob in data.objectives else "")
                ox = csv_scope(ob, _OBJECTIVE_UPPER)
                yield (num_to_str(data.objective_bounds[ox])
                       if ox in data.objective_bounds else "")

    def get_header_comments(self) -> Iterable[str]:
        """
        Get any possible header comments.

        :returns: the header comments
        """
        return ("End Results of Bin Packing Experiments",
                "See the description at the bottom of the file.")

    def get_footer_comments(self) -> Iterable[str]:
        """
        Get any possible footer comments.

        :return: the footer comments
        """
        yield from self.__er.get_footer_comments()
        yield ""
        p: Final[str | None] = self.scope
        if self.__bin_bounds:
            for bb in self.__bin_bounds:
                yield (f"{csv_scope(p, bb)} is a lower bound "
                       f"for the number of bins.")
        if self.__objectives:
            for obb in self.__objectives:
                ob: str = csv_scope(p, obb)
                ox: str = csv_scope(ob, _OBJECTIVE_LOWER)
                yield f"{ox}: a lower bound of the {ob} objective function."
                yield (f"{ob}: one of the possible objective functions for "
                       "the two-dimensional bin packing problem.")
                ox = csv_scope(ob, _OBJECTIVE_UPPER)
                yield f"{ox}: an upper bound of the {ob} objective function."

    def get_footer_bottom_comments(self) -> Iterable[str]:
        """
        Get the bottom footer comments.

        :param dest: the destination
        """
        yield from motipyapps_footer_bottom_comments(
            self, "The packing data is assembled using module "
                  "moptipyapps.binpacking2d.packing_statistics.")
        yield from ErCsvWriter.get_footer_bottom_comments(self.__er)


class CsvReader:
    """A class for CSV reading of :class:`PackingResult` instances."""

    def __init__(self, columns: dict[str, int]) -> None:
        """
        Create a CSV parser for :class:`EndResult`.

        :param columns: the columns
        """
        super().__init__()
        #: the end result csv reader
        self.__er: Final[ErCsvReader] = ErCsvReader(columns)
        #: the index of the n-items column
        self.__idx_n_items: Final[int] = csv_column(columns, KEY_N_ITEMS)
        #: the index of the n different items column
        self.__idx_n_different: Final[int] = csv_column(
            columns, KEY_N_DIFFERENT_ITEMS)
        #: the index of the bin width column
        self.__idx_bin_width: Final[int] = csv_column(columns, KEY_BIN_WIDTH)
        #: the index of the bin height  column
        self.__idx_bin_height: Final[int] = csv_column(
            columns, KEY_BIN_HEIGHT)
        #: the indices for the objective bounds
        self.__bin_bounds: Final[tuple[tuple[str, int], ...]] = \
            csv_select_scope(
                lambda x: tuple(sorted(((k, v) for k, v in x.items()))),
                columns, LOWER_BOUNDS_BIN_COUNT)
        #: the objective bounds
        self.__objective_bounds: Final[tuple[tuple[str, int], ...]] = \
            csv_select_scope(
                lambda x: tuple(sorted(((k, v) for k, v in x.items()))),
                columns, None,
                skip_orig_key=lambda s: not str.endswith(
                    s, (_OBJECTIVE_LOWER, _OBJECTIVE_UPPER)))
        n_bounds: Final[int] = tuple.__len__(self.__objective_bounds)
        if n_bounds <= 0:
            raise ValueError("No objective function bounds found?")
        if (n_bounds & 1) != 0:
            raise ValueError(f"Number of bounds {n_bounds} should be even.")
        #: the parsers for the objective values
        self.__objectives: Final[tuple[tuple[str, int], ...]] = \
            tuple((ss, csv_column(columns, ss))
                  for ss in sorted({s[0] for s in (str.split(
                      kk[0], SCOPE_SEPARATOR) for kk in
                      self.__objective_bounds) if (list.__len__(s) > 1)
                      and (str.__len__(s[0]) > 0)}))
        n_objectives: Final[int] = tuple.__len__(self.__objectives)
        if n_objectives <= 0:
            raise ValueError("No objectives found?")
        if (2 * n_objectives) != n_bounds:
            raise ValueError(
                f"Number {n_objectives} of objectives "
                f"inconsistent with number {n_bounds} of bounds.")

    def parse_row(self, data: list[str]) -> PackingResult:
        """
        Parse a row of data.

        :param data: the data row
        :return: the end result statistics
        """
        return PackingResult(
            self.__er.parse_row(data),
            int(data[self.__idx_n_items]),
            int(data[self.__idx_n_different]),
            int(data[self.__idx_bin_width]),
            int(data[self.__idx_bin_height]),
            {n: str_to_num(data[i]) for n, i in self.__objectives
             if str.__len__(data[i]) > 0},
            {n: str_to_num(data[i]) for n, i in self.__objective_bounds
             if str.__len__(data[i]) > 0},
            {n: int(data[i]) for n, i in self.__bin_bounds
             if str.__len__(data[i]) > 0})


class __LogParser(LogParser):
    """The internal log parser class."""

    def __init__(self, consumer: Callable[[PackingResult], None],
                 objectives: Iterable[Callable[[Instance], Objective]],
                 bin_bounds: Mapping[str, Callable[[Instance], int]]) -> None:
        """
        Parse a directory recursively to get all packing results.

        :param consumer: the consumer for receiving the results
        :param objectives: the objective function factories
        :param bin_bounds: the bin bounds calculators
        """
        super().__init__()
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        if not isinstance(objectives, Iterable):
            raise type_error(objectives, "objectives", Iterable)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        #: the internal consumer
        self.__consumer: Final[Callable[[PackingResult], None]] = consumer
        #: the objectives holder
        self.__objectives: Final[
            Iterable[Callable[[Instance], Objective]]] = objectives
        #: the bin bounds
        self.__bin_bounds: Final[
            Mapping[str, Callable[[Instance], int]]] = bin_bounds
        #: the internal cache
        self.__cache: Final[Mapping[
            str, tuple[Mapping[str, int], tuple[
                Objective, ...], Mapping[str, int | float]]]] = {}

    def parse_file(self, path: str) -> bool:
        """
        Parse a log file.

        :param path: the path to the log file
        :return: `True`
        """
        self.__consumer(from_single_log(path, self.__objectives,
                                        self.__bin_bounds, self.__cache))
        return True


# Evaluate an experiment from the command line

# Run log files to end results if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipyapps_argparser(
        __file__,
        "Convert log files for the bin packing experiment to a CSV file.",
        "Re-evaluate all results based on different objective functions.")
    parser.add_argument(
        "source", nargs="?", default="./results",
        help="the location of the experimental results, i.e., the root folder "
             "under which to search for log files", type=Path)
    parser.add_argument(
        "dest", help="the path to the end results CSV file to be created",
        type=Path, nargs="?", default="./evaluation/end_results.txt")
    args: Final[argparse.Namespace] = parser.parse_args()

    packing_results: Final[list[PackingResult]] = []
    from_logs(args.source, packing_results.append)
    to_csv(packing_results, args.dest)



"""
An extended dynamic control problem `Instance` with a model for the dynamics.

An :class:`~moptipyapps.dynamic_control.instance.Instance` is a combination of

- a set of differential equations
  (:mod:`~moptipyapps.dynamic_control.system`) that govern the change
  `D=ds/dt` of the state `s` of a dynamic system (based on its current
  state `s` and the controller output `c(s)`), i.e., `ds/dt=D(s,c(s))` and
- a blueprint `c(s, p)` of the controller `c(s)`
  (:mod:`~moptipyapps.dynamic_control.controller`).

The blueprint of the controller is basically a function that can be
parameterized, so it is actually a function `c(s, p)` and the goal of
optimization is to parameterize it in such a way that the figure of merit,
i.e., the objective value (:mod:`~moptipyapps.dynamic_control.objective`)
of the system (usually the sum of squared state values and squared controller
outputs) is minimized. Parameterization means finding good values `p` such
that the above goal is reached. In other words, we want to synthesize a
controller (by finding good values of `p`) in such a way that the state
equations drive the system into a stable state, usually ideally the origin of
the coordinate system.

Now here this :class:`~moptipyapps.dynamic_control.instance.Instance` is
extended to a :class:`~moptipyapps.dynamic_control.system_model.SystemModel`
by adding a parameterized model `M(s, c(s), q)` to the mix. The idea is to
develop the parameterization `q` of the model `M` that can replace the actual
system equations `D`. Such a model receives as input the current system state
vector `s` and the controller output `c(s)` for the state vector `s`. It will
return the differential `D` of the system state, i.e., `ds/dt`. In other
words, a properly constructed model can replace the `equations` parameter in
the ODE integrator :func:`~moptipyapps.dynamic_control.ode.run_ode`. The
input used for training is provided by
:func:`~moptipyapps.dynamic_control.ode.diff_from_ode`.

What we do here is to re-use the same function models as used in controllers
(:mod:`~moptipyapps.dynamic_control.controller`) and learn their
parameterizations from the observed data. If successful, we can  wrap
everything together into a `Callable` and plug it into the system instead of
the original equations.

The thing that :class:`SystemModel` offers is thus a blueprint of the model
`M`. Obviously, we can conceive many different such blueprints. We could have
a linear model, a quadratic model, or maybe neural network (in which case, we
need to decide about the number of layers and layer sizes). So an instance of
this surrogate model based approach has an equation system, a controller
blueprint, and a model blueprint.

An example implementation of the concept of synthesizing models for a dynamic
system in order to synthesize controllers is given as the
:mod:`~moptipyapps.dynamic_control.surrogate_optimizer` algorithm.
Examples for different dynamic systems controllers (which we here also use to
model the systems themselves) are given in package
:mod:`~moptipyapps.dynamic_control.controllers`.

The starting points of the work here were conversations with Prof. Dr. Bernd
NOACK and Guy Yoslan CORNEJO MACEDA of the Harbin Institute of Technology in
Shenzhen, China (哈尔滨工业大学(深圳)).
"""
from typing import Final

from moptipy.utils.logger import KeyValueLogSection
from pycommons.types import type_error

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.instance import Instance
from moptipyapps.dynamic_control.system import System


class SystemModel(Instance):
    """A dynamic system control `Instance` with a system model blueprint."""

    def __init__(self, system: System, controller: Controller,
                 model: Controller) -> None:
        """
        Initialize the system model.

        :param system: the system that we want to model
        :param controller: the controller that we want to use
        :param model: a system model blueprint (also in the shape of a
            controller)
        """
        if not isinstance(model, Controller):
            raise type_error(model, "model", Controller)
        super().__init__(system, controller, model.name)
        in_dims: Final[int] = system.state_dims + controller.control_dims
        if model.state_dims != in_dims:
            raise ValueError(
                f"model.state_dims={model.state_dims}, but system.state_dims"
                f"={system.state_dims} and controller.control_dims="
                f"{controller.control_dims}, so we expected {in_dims} for "
                f"controller {str(controller)!r}, system {str(system)!r}, and"
                f" model {str(model)!r}.")
        if model.control_dims != system.state_dims:
            raise ValueError(
                f"model.control_dims={model.control_dims} must be the same as"
                f" system.state_dims={system.state_dims}, but is not, for "
                f"controller {str(controller)!r}, system {str(system)!r}, and"
                f" model {str(model)!r}.")
        #: the model blueprint that can be trained to hopefully replace the
        #: :attr:`~moptipyapps.dynamic_control.instance.Instance.system` in the
        #: ODE integration / system simulation procedure
        self.model: Final[Controller] = model

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope("model") as scope:
            self.model.log_parameters_to(scope)



"""An example experiment for generating bin packing instances."""

import argparse
from typing import Callable, Final, cast

from moptipy.algorithms.so.vector.cmaes_lib import BiPopCMAES
from moptipy.api.execution import Execution
from moptipy.api.experiment import run_experiment
from pycommons.io.path import Path

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.instgen.errors_and_hardness import (
    ErrorsAndHardness,
)
from moptipyapps.binpacking2d.instgen.problem import Problem
from moptipyapps.shared import moptipyapps_argparser

#: the maximum number of FEs
MAX_FES: Final[int] = 10_000

#: the maximum number of FEs in for the algorithm runs inside the objective
INNER_MAX_FES: Final[int] = 225_000

#: the number of runs of the algorithms inside the objective
INNER_RUNS: Final[int] = 2


def cmaes(problem: Problem) -> Execution:
    """
    Create the basic BiPop-CMA-ES setup.

    :param problem: the problem to solve
    :return: the execution
    """
    return (Execution()
            .set_algorithm(BiPopCMAES(problem.search_space))
            .set_search_space(problem.search_space)
            .set_solution_space(problem.solution_space)
            .set_encoding(problem.encoding)
            .set_objective(ErrorsAndHardness(
                problem.solution_space, INNER_MAX_FES, INNER_RUNS))
            .set_max_fes(MAX_FES)
            .set_log_improvements(True))


#: the instances to use as blueprints
__USE_INSTANCES: Final[tuple[str, ...]] = (
    "beng01", "beng02", "beng03", "beng04", "beng05", "beng06", "beng07",
    "beng08", "beng09", "beng10", "cl01_020_01", "cl01_040_01", "cl01_060_01",
    "cl01_080_01", "cl01_100_01", "cl02_020_01", "cl02_040_01", "cl02_060_01",
    "cl02_080_01", "cl02_100_01", "cl03_020_01", "cl03_040_01", "cl03_060_01",
    "cl03_080_01", "cl03_100_01", "cl04_020_01", "cl04_040_01", "cl04_060_01",
    "cl04_080_01", "cl04_100_01", "cl05_020_01", "cl05_040_01", "cl05_060_01",
    "cl05_080_01", "cl05_100_01", "cl06_020_01", "cl06_040_01", "cl06_060_01",
    "cl06_080_01", "cl06_100_01", "cl07_020_01", "cl07_040_01", "cl07_060_01",
    "cl07_080_01", "cl07_100_01", "cl08_020_01", "cl08_040_01", "cl08_060_01",
    "cl08_080_01", "cl08_100_01", "cl09_020_01", "cl09_040_01", "cl09_060_01",
    "cl09_080_01", "cl09_100_01", "cl10_020_01", "cl10_040_01", "cl10_060_01",
    "cl10_080_01", "cl10_100_01", "cl10_100_10")


def run(base_dir: str) -> None:
    """
    Run the experiment.

    :param base_dir: the base directory
    """
    use_dir: Final[Path] = Path(base_dir)
    use_dir.ensure_dir_exists()

    inst_creators: list[Callable[[], Instance]] = [cast(
        Callable[[], Instance], lambda __s=_s, __t=_t: Problem(__s, __t))
        for _s in __USE_INSTANCES for _t in (0.25, 0.125)]

    run_experiment(
        base_dir=use_dir,
        instances=inst_creators,
        setups=[cmaes],
        n_runs=[1, 2, 3, 5, 7, 11, 13, 17, 20, 23, 29, 30, 31],
        perform_warmup=False,
        perform_pre_warmup=False)


# Run the experiment from the command line
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipyapps_argparser(
        __file__, "2D Bin Packing Instance Generator",
        "Run the 2D Bin Packing Instance Generator experiment.")
    parser.add_argument(
        "dest", help="the directory to store the experimental results under",
        type=Path, nargs="?", default="./results/")
    args: Final[argparse.Namespace] = parser.parse_args()
    run(args.dest)



"""
A quadratic controller.

A quadratic controller is a function where all value of the state vector enter
the computation plainly and squared. The powers of their combinations do not
exceed two, e.g., the controller is a linear combination of A, B, A², AB, and
B² if the state has values A and B. The controller represents the multipliers
for these coefficients.
"""

from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __quadratic_2d_1o(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a quadratic function for 2d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    out[0] = (s0 * params[0]) + (s1 * params[1]) + (s0 * s0 * params[2]) \
        + (s0 * s1 * params[3]) + (s1 * s1 * params[4])


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __quadratic_3d_1o(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a quadratic function for 3d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    out[0] = (s0 * params[0]) + (s1 * params[1]) + (s2 * params[2]) \
        + (s0 * s0 * params[3]) + (s0 * s1 * params[4]) \
        + (s0 * s2 * params[5]) + (s1 * s1 * params[6]) \
        + (s1 * s2 * params[7]) + (s2 * s2 * params[8])


def quadratic(system: System) -> Controller:
    """
    Create a quadratic controller for the given equations object.

    :param system: the equations object
    :return: the quadratic controller
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    name: Final[str] = "quadratic"
    if system.state_dims == 2:
        return Controller(name, 2, 1, 5, __quadratic_2d_1o)
    if system.state_dims == 3:
        return Controller(name, 3, 1, 9, __quadratic_3d_1o)
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
An objective function indirectly minimizing the number of bins in packings.

This objective function computes the number of bins used. Let's call it
`n_bins`. We know the area `bin_area` of a bin as well. We then compute the
minimum area occupied in any bin, `min_area`. Now we return
`(bin_area * (n_bins - 1)) + min_area`.

This function always prefers a packing that has fewer bins over a packing
with more bins duw to the term `bin_area * (n_bins - 1)` and
`bin_area >= min_area` (let us ignore the case where `bin_area == min_area`,
which does not make practical sense). Since `min_area < bin_area` in all
practically relevant cases, the offset `min_area` just distinguishes
packings that have same number of bins. Amongst such packings, those whose
least-occupied bin is closer to being empty are preferred (regardless where
this bin is). The idea is that this will eventually allow us to get rid of
that least-occupied bin in subsequent optimization steps, i.e., to reduce
the number of bins.

This is similar to the :mod:`~moptipyapps.binpacking2d.objectives.\
bin_count_and_small` objective, except that we do not try to expunge the last
bin, but any bin. It has the same lower and upper bound, though.
"""
from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.objectives.bin_count_and_last_small import (
    BinCountAndLastSmall,
)
from moptipyapps.binpacking2d.packing import (
    IDX_BIN,
    IDX_BOTTOM_Y,
    IDX_LEFT_X,
    IDX_RIGHT_X,
    IDX_TOP_Y,
)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def bin_count_and_small(y: np.ndarray, bin_area: int,
                        temp: np.ndarray) -> int:
    """
    Compute the number of bins and the smallest occupied area in any bin.

    We compute the total number of bins minus 1 and multiply it with the
    total area of items. We then add the area of items in the smallest bin.

    :param y: the packing
    :param bin_area: the area of a single bin
    :param temp: a temporary array to hold the current area counters
    :return: the objective value

    >>> tempa = np.empty(10, int)
    >>> bin_count_and_small(np.array([[1, 1, 10, 10, 20, 20],
    ...                               [1, 1, 30, 30, 40, 40],
    ...                               [1, 1, 20, 20, 30, 30]], int),
    ...                              50*50, tempa)
    300
    >>> bin_count_and_small(np.array([[1, 1, 10, 10, 20, 20],
    ...                               [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                               [1, 1, 20, 20, 30, 30]], int),
    ...                              50*50, tempa)
    2600
    >>> bin_count_and_small(np.array([[1, 2, 10, 10, 20, 20],  # bin 2!
    ...                               [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                               [1, 1, 20, 20, 30, 30]], int),
    ...                               50*50, tempa)
    2600
    >>> bin_count_and_small(np.array([[1, 3, 10, 10, 20, 20],  # bin 3!
    ...                               [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                               [1, 1, 20, 20, 30, 30]], int),
    ...                              50*50, tempa)
    5100
    >>> bin_count_and_small(np.array([[1, 3, 10, 10, 50, 50],  # bin 3!
    ...                               [1, 2, 30, 30, 60, 60],  # bin 2!
    ...                               [1, 1, 20, 20, 30, 30]], np.int8),
    ...                              50*50, tempa)
    5100
    """
    total_bins: int = 0  # the current idea of the number of bins
    n_items: Final[int] = len(y)  # the number of rows in the matrix
    temp.fill(0)  # fill temp with zeros

    for i in range(n_items):  # iterate over all packed items
        bin_idx: int = int(y[i, IDX_BIN]) - 1  # get the bin index of the item
        temp[bin_idx] += ((y[i, IDX_RIGHT_X] - y[i, IDX_LEFT_X])
                          * (y[i, IDX_TOP_Y] - y[i, IDX_BOTTOM_Y]))
        total_bins = max(total_bins, bin_idx)
    return (bin_area * total_bins) + temp[0:total_bins + 1].min()


class BinCountAndSmall(BinCountAndLastSmall):
    """Compute the number of bins and the area in the smallest one."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__(instance)
        #: the internal temporary array
        self.__temp: Final[np.ndarray] = np.empty(instance.n_items, int)

    def evaluate(self, x) -> int:
        """
        Evaluate the objective function.

        :param x: the solution
        :return: the bin size and smallest-bin-area factor
        """
        return bin_count_and_small(x, self._bin_size, self.__temp)

    def __str__(self) -> str:
        """
        Get the name of the bins objective function.

        :return: `binCountAndSmall`
        :retval "binCountAndSmall": always
        """
        return "binCountAndSmall"



"""
A surrogate system model-based Optimization approach.

In the real world, we want to synthesize a controller `c(s, p)` that can
drive a dynamic system into a good state. The controller receives as input
the current state `s`, say, from sensor readings. It can also be
parameterized by a vector `p`, imagine `c` to be, for example, an artificial
neural network and then `p` would be its weight vector. The output of `c` will
influence the system in some way. In our example experiments, this is done
by becoming part of the state differential `ds/dt`. Anyway, in the real world,
the controller may steer the rotation speed of a rotor or the angles of rotor
blades or whatever. Now you can imagine that doing real-world experiments is
costly and takes a long time. Everytime we want to test a parameterization
`p`, some form experiment, maybe in a wind tunnel, has to be done.

So it would be beneficial if we could replace the actual experiment by a
simulation. This would mean that we learn a model `M` that can compute the
state change `ds/dt` based on the current state `s` and controller output `c`
at reasonable accuracy. If we had such a computational model, then we could
run the controller optimization process on that model. Once finished, we could
apply and evaluate the best controller that we have discovered in a real
experiment. With the observed behavior of the actual controller system, we may
even be able to update and improve our system model to become more accurate.
So we could alternate between real experiments and optimization runs on the
simulation. Of course, we would always need to do some real experiments at
first to gather the data to obtain our initial model `M`. But if we can get
this to work, then we could probably get much better controllers with fewer
actual experiments.

This here is an algorithm that tries to implement the above pattern.
This algorithm employs three different sub-algorithms:

1. For sampling the initial :attr:`~SurrogateOptimizer.fes_for_warmup` FEs,
   it uses a warm-up algorithm - by default, this is done by
   :mod:`~moptipy.algorithms.random_sampling`.
2. Then, it spends :attr:`~SurrogateOptimizer.fes_for_training` steps on the
   collected data to train the model using the model training algorithm,
   which, by default, is
   :class:`~moptipy.algorithms.so.vector.cmaes_lib.BiPopCMAES`.
3. Then, it spends :attr:`~SurrogateOptimizer.fes_per_model_run` steps
   to train controllers on the model, by default again with a
   :class:`~moptipy.algorithms.so.vector.cmaes_lib.BiPopCMAES`.

For the model and controller optimization, it thus uses by default the
BiPop-CMA-ES offered by `moptipy`
(:class:`~moptipy.algorithms.so.vector.cmaes_lib.BiPopCMAES`).
But it uses two instances of this algorithm, namely one to optimize the
controller parameters and one that optimizes the model parameters.
And, as said, a random sampling method to gather the initial samples.

The idea is that we divide the computational budget into a warmup and a
model-based phase. In the warmup phase, we use CMA-ES to normally optimize
the controller based on the actual simulation of the dynamic system.
However, while doing so, for every time step in the simulation, we collect
three things: The current state vector `s`, the control vector `c`, and the
resulting state differential `ds/dt`. Now if we have such data, we can look
at the dynamic system as a function `D(s, c) = ds/dt`. If we consider the
dynamic system to be such a function and we have collected the vectors
`(s, c)` and `ds/dt`, then we may as well attempt to *learn* this system.
So after the warmup phase, our algorithm does the following: In a loop (for
the rest of the computational budget), it first tries to learn a model `M` of
`D`. Then, it replaces the actual differential equations of the system in ODE
solution approach of the objective function with `M`. In other words, we kick
out the actual system and instead use the learned system model `M`. We replace
the differential equations that describe the system using `M`. We can now
run an entire optimization process on this learned model only, with ODE
integration and all. This optimization process gives us one new solution which
we then evaluate on the real objective function (which costs 1 FE and gives us
a new heap of `(s, c)` and `ds/dt` vectors). With this new data, we again
learn a new and hopefully more accurate model `M`. This process is iterated
until the rest of the computational budget is exhausted.

This approach hopefully allows us to learn a model of a dynamic system while
synthesizing a controller for it. Since we can have infinitely more time to
synthesize the controller on a learned system model compared to an actual
model, this may give us much better results.

The starting points of the work here were conversations with Prof. Dr. Bernd
NOACK and Guy Yoslan CORNEJO MACEDA of the Harbin Institute of Technology in
Shenzhen, China (哈尔滨工业大学(深圳)).
"""


from copy import copy
from gc import collect
from os.path import basename
from typing import Callable, Final

import numba  # type: ignore
import numpy as np
from moptipy.algorithms.random_sampling import RandomSampling
from moptipy.algorithms.so.vector.cmaes_lib import BiPopCMAES
from moptipy.api.algorithm import Algorithm
from moptipy.api.execution import Execution
from moptipy.api.logging import FILE_SUFFIX
from moptipy.api.process import Process
from moptipy.api.subprocesses import for_fes
from moptipy.operators.vectors.op0_uniform import Op0Uniform
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import rand_seed_generate
from numpy.random import Generator
from pycommons.io.path import Path
from pycommons.types import check_int_range, type_error

from moptipyapps.dynamic_control.model_objective import ModelObjective
from moptipyapps.dynamic_control.objective import FigureOfMerit
from moptipyapps.dynamic_control.system import System
from moptipyapps.dynamic_control.system_model import SystemModel


def _nop() -> None:
    """Do absolutely nothing."""


def _bpcmaes(vs: VectorSpace) -> BiPopCMAES:
    """
    Create the Bi-Pop CMA-ES.

    :param vs: the vector space
    :return: the algorithm
    """
    return BiPopCMAES(vs, True)


class SurrogateOptimizer(Algorithm):
    """A surrogate model-based CMA-ES algorithm."""

    def __init__(
            self, system_model: SystemModel, controller_space: VectorSpace,
            objective: FigureOfMerit, fes_for_warmup: int,
            fes_for_training: int | None = None,
            ms_for_training: int | None = None,
            fes_per_model_run: int | None = None,
            ms_per_model_run: int | None = None,
            fancy_logs: bool = False,
            warmup_algorithm: Callable[
                [VectorSpace], Algorithm] =
            lambda v: RandomSampling(Op0Uniform(v)),
            model_training_algorithm: Callable[
                [VectorSpace], Algorithm] = _bpcmaes,
            controller_training_algorithm: Callable[
                [VectorSpace], Algorithm] = _bpcmaes) -> None:
        """
        Initialize the algorithm.

        :param system_model: the system and model
        :param controller_space: the controller space
        :param objective: the figure of merit
        :param fes_for_warmup: the number of objective function evaluations
            (FEs) to be used on the initial stage on the real system for
            warming up
        :param fes_for_training: the number of FEs to be used to train the
            model
        :param ms_for_training: the number of milliseconds for model training
        :param fes_per_model_run: the number of FEs to be applied to each
            optimization run on the model
        :param ms_per_model_run: the number of milliseconds for a model run
        :param fancy_logs: should we perform fancy logging?
        :param warmup_algorithm: the algorithm for sampling the warmup points
        :param model_training_algorithm: the model training algorithm builder
        :param controller_training_algorithm: the controller training
            algorithm builder
        """
        super().__init__()

        if not isinstance(system_model, SystemModel):
            raise type_error(system_model, "system_model", SystemModel)
        if not isinstance(controller_space, VectorSpace):
            raise type_error(
                controller_space, "controller_space", VectorSpace)
        if not isinstance(objective, FigureOfMerit):
            raise type_error(objective, "objective", FigureOfMerit)
        if not isinstance(fancy_logs, bool):
            raise type_error(fancy_logs, "fancy_logs", bool)
        if not callable(warmup_algorithm):
            raise type_error(warmup_algorithm, "warmup_algorithm", call=True)
        if not callable(model_training_algorithm):
            raise type_error(model_training_algorithm,
                             "model_training_algorithm", call=True)
        if not callable(controller_training_algorithm):
            raise type_error(controller_training_algorithm,
                             "controller_training_algorithm", call=True)

        # the controller space
        self.controller_space: Final[VectorSpace] = controller_space
        #: should we do fancy logging?
        self.fancy_logs: Final[bool] = fancy_logs
        #: the number of objective function evaluations to be used for warmup
        self.fes_for_warmup: Final[int] = check_int_range(
            fes_for_warmup, "fes_for_warmup", 1, 1_000_000_000)
        #: the FEs for training the model
        self.fes_for_training: Final[int | None] = None \
            if fes_for_training is None else check_int_range(
            fes_for_training, "fes_for_training", 1, 1_000_000_000)
        #: the ms for training the model
        self.ms_for_training: Final[int | None] = None \
            if ms_for_training is None else check_int_range(
            ms_for_training, "ms_for_training", 1, 1_000_000_000_000)
        if (self.fes_for_training is None) and (self.ms_for_training is None):
            raise ValueError("One of fes_for_training or "
                             "ms_for_training must not be None.")
        #: the FEs for each run on the model
        self.fes_per_model_run: Final[int | None] = None \
            if fes_per_model_run is None else check_int_range(
            fes_per_model_run, "fes_per_model_run", 1, 1_000_000_000)
        #: the ms for each run on the model
        self.ms_per_model_run: Final[int | None] = None \
            if ms_per_model_run is None else check_int_range(
            ms_per_model_run, "ms_per_model_run", 1, 1_000_000_000_000)
        if ((self.fes_per_model_run is None)
                and (self.ms_per_model_run is None)):
            raise ValueError("One of fes_per_model_run or "
                             "ms_per_model_run must not be None.")
        #: the system model
        self.system_model: Final[SystemModel] = system_model
        #: the internal warmup algorithm
        self.__warmup_algorithm: Final[Algorithm] = warmup_algorithm(
            controller_space)
        #: the internal controller training algorithm
        self.__control_algorithm: Final[Algorithm] = (
            controller_training_algorithm(controller_space))
        #: the model parameter space
        self.__model_space: Final[VectorSpace] = \
            system_model.model.parameter_space()
        #: the model training algorithm
        self.__model_training: Final[Algorithm] = model_training_algorithm(
            self.__model_space)
        #: the control objective function reference
        self.__control_objective: Final[FigureOfMerit] = objective
        #: the model objective
        self.__model_objective: Final[ModelObjective] = ModelObjective(
            objective, system_model.model)

    def initialize(self) -> None:
        """Initialize."""
        super().initialize()
        self.__warmup_algorithm.initialize()
        self.__model_training.initialize()
        self.__control_algorithm.initialize()
        self.__model_space.initialize()

    def solve(self, process: Process) -> None:
        """
        Solve the modelling problem.

        This function begins by spending :attr:`fes_for_warmup` objective
        function evaluations (FEs) on the actual problem, i.e., by trying to
        synthesize controllers for the "real" system, using `process` to
        evaluate the controller performance. The objective function passed to
        the constructor (an instance of
        :class:`~moptipyapps.dynamic_control.objective.FigureOfMerit`) must be
        used by `process` as well. This way, during the warm-up phase, we can
        collect tuples of the (system state, controller output) and the
        resulting system differential for each simulated time step. This is
        done via
        :meth:`~moptipyapps.dynamic_control.objective.FigureOfMerit.set_raw`.

        After the warm-up phase, we can obtain these collected data via
        :meth:`~moptipyapps.dynamic_control.objective.FigureOfMerit.\
get_differentials`. The data is then used to train a model via the model
        objective function
        :mod:`~moptipyapps.dynamic_control.model_objective`. The system model
        is again basically a
        :class:`~moptipyapps.dynamic_control.controller.Controller` which is
        parameterized appropriately. For this, we use a CMA-ES algorithm for
        :attr:`fes_for_training` FEs.

        Once the model training is completed, we switch the objective function
        to use the model instead of the actual system for evaluating
        controllers, which is done via :meth:`~moptipyapps.dynamic_control.\
objective.FigureOfMerit.set_model`. We then train a completely new controller
        on the model objective function. Notice that now, the actual system is
        not involved at all. We do this again using a CMA-ES algorithm for
        :attr:`fes_per_model_run` FEs.

        After training the controller, we can evaluate it on the real system
        using the :meth:`~moptipy.api.process.Process.evaluate` method
        of the actual `process` (after switching back to the real model via
        :meth:`~moptipyapps.dynamic_control.objective.FigureOfMerit.set_raw`).
        This nets us a) the actual controller performance and b) a new set of
        (system state, controller output) + system state differential tuples.

        Since we now have more data, we can go back and train a new system
        model and then use this model for another model-based optimization
        run. And so on, and so on. Until the budget is exhausted.

        :param process: the original optimization process, which must use
            the `objective` function (an instance of
            :class:`~moptipyapps.dynamic_control.objective.FigureOfMerit`) as
            its objective function.
        """
        # First, we set up the local variables and fast calls.
        should_terminate: Final[Callable[[], bool]] = process.should_terminate
        model_space: Final[VectorSpace] = self.__model_space
        model_objective: Final[ModelObjective] = self.__model_objective
        model = model_space.create()
        model_equations: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] =\
            self.system_model.model.controller
        raw: Final[FigureOfMerit] = self.__control_objective
        random: Final[Generator] = process.get_random()

        training_execute: Final[Execution] = (
            Execution().set_solution_space(model_space)
            .set_algorithm(self.__model_training)
            .set_objective(model_objective))
        if self.fes_for_training is not None:
            training_execute.set_max_fes(self.fes_for_training)
        if self.ms_for_training is not None:
            training_execute.set_max_time_millis(self.ms_for_training)

        on_model_execute: Final[Execution] = \
            (Execution().set_solution_space(self.controller_space)
             .set_objective(raw)
             .set_algorithm(self.__control_algorithm))
        if self.fes_per_model_run is not None:
            on_model_execute.set_max_fes(self.fes_per_model_run)
        if self.ms_per_model_run is not None:
            on_model_execute.set_max_time_millis(self.ms_per_model_run)

        result: Final[np.ndarray] = self.controller_space.create()
        orig_init: Callable = raw.initialize

# Get a log dir if logging is enabled and set up all the logging information.
        log_dir_name: str | None = process.get_log_basename() \
            if self.fancy_logs else None
        model_training_dir: Path | None = None
        model_training_log_name: str | None = None
        models_dir: Path | None = None
        models_name: str | None = None
        tempsys: System | None = None
        ctrl_dir: Path | None = None
        ctrl_log_name: str | None = None
        controllers_dir: Path | None = None
        controllers_name: str | None = None
        if log_dir_name is not None:
            log_dir: Final[Path] = Path(log_dir_name)
            log_dir.ensure_dir_exists()
            prefix: str = "modelTraining"
            model_training_dir = log_dir.resolve_inside(prefix)
            model_training_dir.ensure_dir_exists()
            base_name: Final[str] = basename(log_dir_name)
            model_training_log_name = f"{base_name}_{prefix}_"
            training_execute.set_log_improvements(True)
            prefix = "controllerOnModel"
            ctrl_dir = log_dir.resolve_inside(prefix)
            ctrl_dir.ensure_dir_exists()
            ctrl_log_name = f"{base_name}_{prefix}_"
            on_model_execute.set_log_improvements(True)
            prefix = "model"
            models_dir = log_dir.resolve_inside(prefix)
            models_dir.ensure_dir_exists()
            tempsys = copy(self.system_model.system)
            models_name = f"{tempsys.name}_{prefix}_"
            prefix = "controllersOnReal"
            controllers_dir = log_dir.resolve_inside(prefix)
            controllers_dir.ensure_dir_exists()
            controllers_name = f"{base_name}_{prefix}_"

# Now we do the setup run that creates some basic results and
# gathers the initial information for modelling the system.
        with for_fes(process, self.fes_for_warmup) as prc:
            self.__warmup_algorithm.solve(prc)

        while not should_terminate():  # until budget exhausted
            consumed_fes: int = process.get_consumed_fes()

            # We now train a model on the data that was gathered.
            training_execute.set_rand_seed(rand_seed_generate(random))
            if model_training_dir is not None:
                training_execute.set_log_file(
                    model_training_dir.resolve_inside(
                        f"{model_training_log_name}{consumed_fes}"
                        f"{FILE_SUFFIX}"))
            model_objective.begin()  # get the collected data
            with training_execute.execute() as sub:  # train model
                sub.get_copy_of_best_y(model)  # get best model
            model_objective.end()  # dispose the collected data

            setattr(raw, "initialize", _nop)  # prevent resetting to "raw"

# The trained model is wrapped into an equation function that can be passed to
# the ODE integrator.
            @numba.njit(cache=False, inline="always", fastmath=True,
                        boundscheck=False)
            def __new_model(state: np.ndarray, time: float,
                            control: np.ndarray, out: np.ndarray,
                            _params=model, _eq=model_equations) -> None:
                _eq(np.hstack((state, control)), time, _params, out)

            setattr(__new_model, "modelParameters", model)  # see objective

            if tempsys is not None:  # plot the model behavior
                tempsys.equations = __new_model  # type: ignore
                setattr(tempsys, "name", f"{models_name}{consumed_fes}")
                tempsys.describe_system_without_control(models_dir)

            collect()  # now we collect all garbage ... there should be much

# OK, now that we got the model, we can perform the model optimization run.
            raw.set_model(__new_model)  # switch to use the model
            on_model_execute.set_rand_seed(rand_seed_generate(random))
            if ctrl_dir is not None:
                on_model_execute.set_log_file(ctrl_dir.resolve_inside(
                    f"{ctrl_log_name}{consumed_fes}{FILE_SUFFIX}"))
            with on_model_execute.execute() as ome:
                ome.get_copy_of_best_y(result)  # get best controller
            raw.set_raw()  # switch to the actual problem and data collection
            setattr(raw, "initialize", orig_init)  # allow resetting to "raw"

            if tempsys is not None:  # plot the controller on that model
                setattr(tempsys, "name", f"{models_name}{consumed_fes}")
                tempsys.describe_system(
                    f"{models_name}{consumed_fes}",
                    self.system_model.controller.controller, result,
                    f"{models_name}{consumed_fes}_synthesized_controller",
                    models_dir)

# Finally, we re-evaluate the result that we got from the model run on the
# actual objective function.
            process.evaluate(result)  # get the real objective value

# plot the actual behavior
            if controllers_dir is not None:
                self.system_model.system.describe_system(
                    f"{self.system_model.system}_{consumed_fes}",
                    self.system_model.controller.controller,
                    result, f"{controllers_name}{consumed_fes}",
                    controllers_dir)

            collect()  # now we collect all garbage ... there should be much

    def __str__(self):
        """
        Get the name of the algorithm.

        :return: the algorithm name
        """
        s: str = (f"{self.__warmup_algorithm}_{self.__control_algorithm}_"
                  f"{self.__model_training}_{self.fes_for_warmup}_tn")
        if self.fes_for_training is not None:
            s = f"{s}_{self.fes_for_training}fes"
        if self.ms_for_training is not None:
            s = f"{s}_{self.ms_for_training}ms"
        s = f"{s}_sy"
        if self.fes_per_model_run is not None:
            s = f"{s}_{self.fes_per_model_run}fes"
        if self.ms_per_model_run is not None:
            s = f"{s}_{self.ms_per_model_run}ms"
        return s

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("fesForWarmup", self.fes_for_warmup)
        logger.key_value("fesForTraining", self.fes_for_training)
        logger.key_value("fesPerModelRun", self.fes_per_model_run)
        logger.key_value("fancyLogs", self.fancy_logs)
        with logger.scope("controlX") as cx:
            self.controller_space.log_parameters_to(cx)
        with logger.scope("warmupA") as wa:
            self.__warmup_algorithm.log_parameters_to(wa)
        with logger.scope("controlA") as ca:
            self.__control_algorithm.log_parameters_to(ca)
        with logger.scope("controlF") as cf:
            self.__control_objective.log_parameters_to(cf)
        with logger.scope("modelX") as mx:
            self.__model_space.log_parameters_to(mx)
        with logger.scope("modelF") as mf:
            self.__model_objective.log_parameters_to(mf)
        with logger.scope("modelA") as ma:
            self.__model_training.log_parameters_to(ma)



"""
A permutation-with-repetition-based encoding based on games.

A point in the search space is a permutation (potentially with repetitions)
that can be translated to a :class:`~moptipyapps.ttp.game_plan.GamePlan`.
Each value `v` in the permutation represents a game to be played by two of
the `n` teams. There are `n(n-1)` possible games between `n` teams,
distinguishing home and away teams. Given a value `v` from `0..n(n-1)-1`,
we can get the zero-based index of the home team as
`home_idx = (game // (n - 1)) % n`. The away index is computed in two steps,
first we set `away_idx = game % (n - 1)` and if `away_idx >= home_idx`, we
do `away_idx = away_idy + 1`. (Because a team can never play against itself,
the situation that `home_idx == away_idx` does not need to be represented, so
we can "skip" over this possible value by doing the `away_idx = away_idy + 1`
and thus get a more "compact" numeric range for the permutation elements.)

A game schedule for any round-robin tournament with any given number of rounds
can then be represented as permutation (potentially with repetitions) of these
game values. In the decoding procedure, it is processed from beginning to end
each game is then placed into the earliest slot not already occupied by
another game. In other words, it is placed at the earliest day at which both
involved teams do not yet have other games. If no such slot is available, this
game is not placed at all. In this case, there will be some zeros in the game
plan after the encoding. No other constraint is considered at this stage.

In other words, this encoding may produce game plans that violate constraints.
It does not care about the streak length constraints.
It does not ensure that each team always has a game.
Therefore, it should only be used in conjunction with objective functions that
force the search towards feasible solutions, such as the
:mod:`~moptipyapps.ttp.errors` objective.
"""


from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.encoding import Encoding
from moptipy.spaces.permutations import Permutations
from pycommons.types import check_int_range

from moptipyapps.ttp.instance import Instance


def search_space_for_n_and_rounds(n: int, rounds: int) -> Permutations:
    """
    Create a proper search space for the given number of teams and rounds.

    If the instance prescribes a double-round robin tournament, then this
    is just the :meth:`~moptipy.spaces.permutations.Permutations.standard`
    permutations set. Otherwise, it will be a permutation where some
    elements are omitted (for
    :attr:`~moptipyapps.ttp.instance.Instance.rounds` == 1) or duplicated
    (if :attr:`~moptipyapps.ttp.instance.Instance.rounds` > 2).

    If an odd number of rounds is played, then it is not possible that all
    teams have the same number of games at home and away. Then, the
    permutation is generated such that, if the highest numbers of games at
    home for any team is `k`, no other team has less than `k - 1` games at
    home. If the number of rounds is even, then all teams will have the
    same number of home and away games, that is, the number of teams
    divided by two and multiplied by the number of rounds.

    :param n: the number of teams
    :param rounds: the number of rounds
    :return: the search space

    >>> ";".join(map(str, search_space_for_n_and_rounds(2, 2).blueprint))
    '0;1'
    >>> ";".join(map(str, search_space_for_n_and_rounds(2, 3).blueprint))
    '0;1;1'
    >>> ";".join(map(str, search_space_for_n_and_rounds(2, 4).blueprint))
    '0;0;1;1'
    >>> ";".join(map(str, search_space_for_n_and_rounds(2, 5).blueprint))
    '0;0;1;1;1'
    >>> ";".join(map(str, search_space_for_n_and_rounds(3, 1).blueprint))
    '1;2;5'
    >>> ";".join(map(str, search_space_for_n_and_rounds(3, 2).blueprint))
    '0;1;2;3;4;5'
    >>> ";".join(map(str, search_space_for_n_and_rounds(3, 3).blueprint))
    '0;1;1;2;2;3;4;5;5'
    >>> ";".join(map(str, search_space_for_n_and_rounds(4, 1).blueprint))
    '1;2;3;7;8;10'
    >>> ";".join(map(str, search_space_for_n_and_rounds(4, 2).blueprint))
    '0;1;2;3;4;5;6;7;8;9;10;11'
    >>> ";".join(map(str, search_space_for_n_and_rounds(4, 3).blueprint))
    '0;1;1;2;2;3;3;4;5;6;7;7;8;8;9;10;10;11'
    >>> ";".join(map(str, search_space_for_n_and_rounds(4, 4).blueprint))
    '0;0;1;1;2;2;3;3;4;4;5;5;6;6;7;7;8;8;9;9;10;10;11;11'
    >>> ";".join(map(str, search_space_for_n_and_rounds(4, 5).blueprint))
    '0;0;1;1;1;2;2;2;3;3;3;4;4;5;5;6;6;7;7;7;8;8;8;9;9;10;10;10;11;11'
    >>> ";".join(map(str, search_space_for_n_and_rounds(5, 1).blueprint))
    '1;2;4;7;9;10;13;15;16;18'
    >>> ";".join(map(str, search_space_for_n_and_rounds(5, 2).blueprint))
    '0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19'
    >>> ";".join(map(str, search_space_for_n_and_rounds(5, 3).blueprint))
    '0;1;1;2;2;3;4;4;5;6;7;7;8;9;9;10;10;11;12;13;13;14;15;15;16;16;17;18;\
18;19'
    """
    check_int_range(n, "n", 2, 100000)
    check_int_range(n, "rounds", 1, 100000)
    div: Final[int] = n - 1
    games: Final[list[int]] = []
    order: bool = False
    for r in range(rounds):
        # If we have an odd round of games, the very last round needs
        # to be treated differently to ensure that the home-away games
        # distribution is fair.
        normal: bool = (r < (rounds - 1)) or ((rounds % 2) == 0)
        for i in range(n):  # for each city
            for j in range(i):  # for each other city
                order = ((r % 2) == 0) if normal else (not order)
                m1 = i if order else j  # determine home city
                m2 = j if order else i  # determine away city
                if m2 > m1:
                    m2 -= 1
                games.append(m1 * div + m2)  # add encoded game tuple
    games.sort()
    return Permutations(games)  # create permutations with repetition


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def map_games(x: np.ndarray, y: np.ndarray) -> None:
    """
    Translate a permutation of games to a game plan.

    This is a straightforward decoding that places the games into the map one
    by one. Each game is placed at the earliest slot in which it can be
    placed. If a game cannot be placed, it is ignored. This will lead to many
    errors, which can be counted via the :mod:`~moptipyapps.ttp.errors`
    objective.

    :param x: the source permutation
    :param y: the destination game plan

    >>> from moptipy.utils.nputils import int_range_to_dtype
    >>> teams = 2
    >>> rounds = 2
    >>> perm = search_space_for_n_and_rounds(teams, rounds).blueprint
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ 2 -1]
     [-2  1]]
    >>> teams = 4
    >>> rounds = 2
    >>> perm = search_space_for_n_and_rounds(teams, rounds).blueprint
    >>> dest = np.empty((rounds * (teams - 1), teams),
    ...                 int_range_to_dtype(-teams, teams))
    >>> map_games(perm, dest)
    >>> print(dest)
    [[ 2 -1  4 -3]
     [ 3  4 -1 -2]
     [ 4  3 -2 -1]
     [-2  1 -4  3]
     [-3 -4  1  2]
     [-4 -3  2  1]]
    """
    y.fill(0)  # first zero the output matrix
    days, n = y.shape  # the number of days and teams to be scheduled
    div: Final[int] = n - 1  # the divisor for permutation values -> teams

    for game in x:
        home_idx: int = (game // div) % n  # home idx is in 0..n-1
        away_idx: int = game % div  # away index in 0..n-2
        if away_idx >= home_idx:  # "A vs. A" games impossible
            away_idx += 1  # away index in 0..n-1, but != home_idx

        for day in range(days):  # iterate over all possible rows for game
            if (y[day, home_idx] != 0) or (y[day, away_idx] != 0):
                continue  # day already blocked
            y[day, home_idx] = away_idx + 1
            y[day, away_idx] = -(home_idx + 1)
            break


class GameEncoding(Encoding):
    """An encoding that transforms strings of games to game plans."""

    def __init__(self, instance: Instance) -> None:
        """
        Create the game-based encoding.

        :param instance: the instance
        """
        super().__init__()
        #: the instance
        self.instance: Final[Instance] = instance
        self.decode = map_games  # type: ignore

    def search_space(self) -> Permutations:
        """
        Create a proper search space for this game-based encoding.

        The search space is a set of :mod:`~moptipy.spaces.permutations` that
        represents all the games that can take place in the tournament.
        Depending on the number of
        :attr:`~moptipyapps.ttp.instance.Instance.rounds` in the tournament,
        some games may appear multiple times. Home and away games are
        distributed in a fair and deterministic mannner between the teams.

        :return: the search space

        >>> inst = Instance.from_resource("circ4")
        >>> inst.n_cities
        4
        >>> inst.rounds
        2
        >>> ";".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0;1;2;3;4;5;6;7;8;9;10;11'
        >>> inst = Instance(inst.name, inst, inst.teams, inst.rounds,
        ...                 inst.home_streak_min, inst.home_streak_max,
        ...                 inst.away_streak_min, inst.away_streak_max,
        ...                 inst.separation_min, inst.separation_max)
        >>> inst.rounds = 1  # modify number of rounds for copied instance
        >>> ";".join(map(str, GameEncoding(inst).search_space().blueprint))
        '1;2;3;7;8;10'
        >>> inst.rounds = 3  # modify number of rounds for copied instance
        >>> ";".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0;1;1;2;2;3;3;4;5;6;7;7;8;8;9;10;10;11'
        >>> inst.rounds = 4  # modify number of rounds for copied instance
        >>> ";".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0;0;1;1;2;2;3;3;4;4;5;5;6;6;7;7;8;8;9;9;10;10;11;11'
        >>> inst.rounds = 5  # modify number of rounds for copied instance
        >>> ";".join(map(str, GameEncoding(inst).search_space().blueprint))
        '0;0;1;1;1;2;2;2;3;3;3;4;4;5;5;6;6;7;7;7;8;8;8;9;9;10;10;10;11;11'
        """
        return search_space_for_n_and_rounds(
            self.instance.n_cities, self.instance.rounds)



"""An encoding that is inspired by a given instance."""
from typing import Final

from moptipy.api.space import Space
from moptipy.utils.logger import KeyValueLogSection
from pycommons.types import check_int_range, type_error

from moptipyapps.binpacking2d.instance import IDX_HEIGHT, IDX_WIDTH, Instance


class InstanceSpace(Space):
    """An space structure for the instance generation problem and space."""

    def __init__(self, source: Instance) -> None:
        """
        Create the space structure for the instance generation problem.

        :param source: the source instances whose date we take as an example
        """
        super().__init__()
        #: the instance name
        self.inst_name: Final[str] = str.strip(source.name) + "n"
        #: the target number of unique items
        self.n_different_items: Final[int] = check_int_range(
            source.n_different_items, "n_different_items",
            1, 100_000)
        #: the target number of items (including repetitions)
        self.n_items: Final[int] = check_int_range(
            source.n_items, "n_items", source.n_different_items,
            1_000_000_000)
        #: the bin width
        self.bin_width: Final[int] = check_int_range(
            source.bin_width, "bin_width", 1, 1_000_000_000)
        #: the bin height
        self.bin_height: Final[int] = check_int_range(
            source.bin_height, "bin_height", 1, 1_000_000_000)
        #: the minimum number of bins that this instance requires
        self.min_bins: Final[int] = check_int_range(
            min(source.lower_bound_bins, self.n_items), "min_bins",
            1, 1_000_000_000)
        #: the minimum item width
        self.item_width_min: Final[int] = check_int_range(int(min(
            source[:, IDX_WIDTH])), "item_width_min", 1, self.bin_width)
        #: the maximum item width
        self.item_width_max: Final[int] = check_int_range(int(max(
            source[:, IDX_WIDTH])), "item_width_max", self.item_width_min,
            self.bin_width)
        #: the minimum item height
        self.item_height_min: Final[int] = check_int_range(int(min(
            source[:, IDX_HEIGHT])), "item_height_min", 1, self.bin_height)
        #: the maximum item height
        self.item_height_max: Final[int] = check_int_range(int(max(
            source[:, IDX_HEIGHT])), "item_height_max", self.item_height_min,
            self.bin_height)
        #: the total item area
        self.total_item_area: Final[int] = check_int_range(
            source.total_item_area, "total_item_area", 1, 1_000_000_000)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this instance.

        :param logger: the logger
        """
        super().log_parameters_to(logger)
        logger.key_value("instName", self.inst_name)
        logger.key_value("nItems", self.n_items)
        logger.key_value("nDifferentItems", self.n_different_items)
        logger.key_value("binWidth", self.bin_width)
        logger.key_value("binHeight", self.bin_height)
        logger.key_value("minBins", self.min_bins)
        logger.key_value("itemWidthMin", self.item_width_min)
        logger.key_value("itemWidthMax", self.item_width_max)
        logger.key_value("itemHeightMin", self.item_height_min)
        logger.key_value("itemHeightMax", self.item_height_max)
        logger.key_value("totalItemArea", self.total_item_area)

    def create(self) -> list[Instance]:
        """
        Generate a list for receiving an instance.

        :return: the new instance list
        """
        return []

    def copy(self, dest: list[Instance], source: list[Instance]) -> None:
        """
        Copy the instance list.

        :param dest: the destination instance list
        :param source: the source instance list
        """
        dest.clear()
        dest.extend(source)

    def to_str(self, x: list[Instance]) -> str:  # +book
        """
        Convert an instance list to a string.

        :param x: the instance list
        :return: the string representation of x
        """
        return x[0].to_compact_str()

    def from_str(self, text: str) -> list[Instance]:
        """
        Convert an instance string to a list with an instance.

        :param text: the input string
        :return: the element in the space corresponding to `text`
        """
        return [Instance.from_compact_str(text)]

    def is_equal(self, x1: list[Instance], x2: list[Instance]) -> bool:
        """
        Check if the contents of two instances of the data structure are equal.

        :param x1: the first instance
        :param x2: the second instance
        :return: `True` if the contents are equal, `False` otherwise
        """
        return x1 == x2

    def validate(self, x: list[Instance]) -> None:
        """
        Check whether a given point in the space is valid.

        :param x: the point
        :raises TypeError: if the point `x` (or one of its elements, if
            applicable) has the wrong data type
        :raises ValueError: if the point `x` is invalid and/or simply is not
            an element of this space
        """
        if list.__len__(x) != 1:
            raise ValueError("There must be exactly one instance in x.")
        inst: Instance = x[0]
        if not isinstance(inst, Instance):
            raise type_error(inst, "x[0]", Instance)
        if inst.name != self.inst_name:
            raise ValueError(
                f"instance name {inst.name!r} != {self.inst_name!r}.")
        if inst.bin_width != self.bin_width:
            raise ValueError(
                f"instance bin width={inst.bin_width} != {self.bin_width}.")
        if inst.bin_height != self.bin_height:
            raise ValueError(
                "instance bin height="
                f"{inst.bin_height} != {self.bin_height}.")
        if inst.n_items != self.n_items:
            raise ValueError(
                f"instance n_items={inst.n_items} != {self.n_items}.")

    def n_points(self) -> int:
        """
        Get the approximate number of different elements in the space.

        :return: the approximate scale of the space
        """
        return (self.bin_height * self.bin_width) ** self.n_items



"""Test all the links in the project's *.md files."""
# noinspection PyPackageRequirements
from pycommons.dev.tests.links_in_md import check_links_in_md
from pycommons.io.path import file_path
from pycommons.processes.caller import is_ci_run


def test_all_links_in_readme_md() -> None:
    """Test all the links in the README.md file."""
    if is_ci_run():
        check_links_in_md(file_path(__file__).up(2).resolve_inside(
            "README.md"))


def test_all_links_in_contributing_md() -> None:
    """Test all the links in the CONTRIBUTING.md file."""
    if is_ci_run():
        check_links_in_md(file_path(__file__).up(2).resolve_inside(
            "CONTRIBUTING.md"))


def test_all_links_in_security_md() -> None:
    """Test all the links in the SECURITY.md file."""
    if is_ci_run():
        check_links_in_md(file_path(__file__).up(2).resolve_inside(
            "SECURITY.md"))


def test_all_links_in_license() -> None:
    """Test all the links in the LICENSE file."""
    if is_ci_run():
        check_links_in_md(file_path(__file__).up(2).resolve_inside(
            "LICENSE"))



"""
Different encodings for the two-dimensional bin packing problem.

The following encodings are implemented:

- The improved bottom-left encoding 1,
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_1`, processes a
  permutation of objects from beginning to end and places the objects into
  the last bin (according to the improved-bottom-left method) until that bin
  is full and then begins to put them into the next bin.
- The improved bottom-left encoding 2,
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_2`, processes a
  permutation of objects from beginning to end and places the objects into
  the first bin into which they fit (according to the improved-bottom-left
  method). It is slower than
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_1` but its results
  are never worse for any permutation and better for several.

Important initial work on this code has been contributed by Mr. Rui ZHAO
(赵睿), <zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥大学) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""



"""
An example experiment for dynamic control.

In this experiment, we try to synthesize (i.e., parameterize) controllers
(:mod:`~moptipyapps.dynamic_control.controller`) that steer a dynamic system
(:mod:`~moptipyapps.dynamic_control.system`) into a state by using a figure
of merit (:mod:`~moptipyapps.dynamic_control.objective`) which minimizes both
the squared system state and controller effort.

A model-based experiment variant is given in
:mod:`~moptipyapps.dynamic_control.experiment_surrogate`.
"""

import argparse
from os.path import basename, dirname
from typing import Any, Callable, Final, Iterable, cast

import numpy as np
from moptipy.algorithms.so.vector.cmaes_lib import BiPopCMAES
from moptipy.api.execution import Execution
from moptipy.api.experiment import run_experiment
from moptipy.api.process import Process
from pycommons.io.path import Path, directory_path

from moptipyapps.dynamic_control.controllers.ann import anns
from moptipyapps.dynamic_control.controllers.cubic import cubic
from moptipyapps.dynamic_control.controllers.linear import linear
from moptipyapps.dynamic_control.controllers.min_ann import min_anns
from moptipyapps.dynamic_control.controllers.partially_linear import (
    partially_linear,
)
from moptipyapps.dynamic_control.controllers.peaks import peaks
from moptipyapps.dynamic_control.controllers.predefined import predefined
from moptipyapps.dynamic_control.controllers.quadratic import quadratic
from moptipyapps.dynamic_control.instance import Instance
from moptipyapps.dynamic_control.objective import FigureOfMeritLE
from moptipyapps.dynamic_control.systems.lorenz import LORENZ_111
from moptipyapps.dynamic_control.systems.stuart_landau import STUART_LANDAU_111
from moptipyapps.shared import moptipyapps_argparser


def make_instances() -> Iterable[Callable[[], Instance]]:
    """
    Create the instances to be used in the dynamic control experiment.

    :return: the instances to be used in the dynamic control experiment.
    """
    res: list[Callable[[], Instance]] = []
    for system in (STUART_LANDAU_111, LORENZ_111):
        controllers = [linear(system), quadratic(system), cubic(system)]
        controllers.extend(anns(system))
        controllers.extend(min_anns(system))
        controllers.extend(partially_linear(system))
        controllers.extend(predefined(system))
        controllers.extend(peaks(system))
        res.extend(cast(
            Callable[[], Instance], lambda _s=system, _c=controller:
            Instance(_s, _c)) for controller in controllers)
    return res


def base_setup(instance: Instance) -> Execution:
    """
    Create the basic setup.

    :param instance: the instance to use
    :return: the basic execution
    """
    return Execution().set_max_fes(512).set_log_all_fes(True)\
        .set_objective(FigureOfMeritLE(instance))


def cmaes(instance: Instance) -> Execution:
    """
    Create the Bi-Pop-CMA-ES setup.

    :param instance: the problem instance
    :return: the setup
    """
    space = instance.controller.parameter_space()
    return base_setup(instance).set_solution_space(space)\
        .set_algorithm(BiPopCMAES(space))


def on_completion(instance: Any, log_file: Path, process: Process) -> None:
    """
    Plot the corresponding figures and print the objective value components.

    :param instance: the problem instance
    :param log_file: the log file
    :param process: the process
    """
    inst: Final[Instance] = cast(Instance, instance)
    dest_dir: Final[Path] = directory_path(dirname(log_file))
    base_name: str = basename(log_file)
    base_name = base_name[:base_name.rindex(".")]
    result: np.ndarray = cast(np.ndarray, process.create())
    process.get_copy_of_best_x(result)
    j: Final[float] = process.get_best_f()
    inst.describe_parameterization(f"F = {j}", result, base_name, dest_dir)


def run(base_dir: str, n_runs: int = 5) -> None:
    """
    Run the experiment.

    :param base_dir: the base directory
    :param n_runs: the number of runs
    """
    use_dir: Final[Path] = Path(base_dir)
    use_dir.ensure_dir_exists()
    instances: Final[Iterable[Callable[[], Instance]]] = make_instances()
    for maker in instances:
        inst: Instance = maker()
        inst.system.describe_system_without_control(use_dir, True)
        inst.system.plot_points(use_dir, True)

    run_experiment(
        base_dir=use_dir,
        instances=instances,
        setups=[cmaes],
        n_runs=n_runs,
        perform_warmup=False,
        perform_pre_warmup=False,
        on_completion=on_completion)


# Run the experiment from the command line
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipyapps_argparser(
        __file__, "Dynamic Control", "Run the dynamic control experiment.")
    parser.add_argument(
        "dest", help="the directory to store the experimental results under",
        type=Path, nargs="?", default="./results/")
    args: Final[argparse.Namespace] = parser.parse_args()
    run(args.dest)



"""
Different objective functions for two-dimensional bin packing.

The following objective functions are implemented:

- :mod:`~moptipyapps.binpacking2d.objectives.bin_count` returns the number
  of bins occupied by a given packing.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_empty` returns
  a combination of the number of bins occupied by a given packing and the
  fewest number of objects located in any bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_empty`
  returns a combination of the number of bins occupied by a given packing and
  the number of objects located in the last bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_small` returns
  a combination of the number of bins occupied by a given packing and the
  smallest area occupied by objects in any bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_small` returns
  a combination of the number of bins occupied by a given packing and the
  area occupied by the objects in the last bin.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_lowest_skyline`
  returns a combination of the number of bins occupied by a given packing and
  the smallest area under the skyline in any bin, where the "skyline" is the
  upper border of the space occupied by objects.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_skyline`
  returns a combination of the number of bins occupied by a given packing and
  the smallest area under the skyline in the last bin, where the "skyline" is
  the upper border of the space occupied by objects.

Important initial work on this code has been contributed by Mr. Rui ZHAO
(赵睿), <zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥大学) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""



"""An instance of the instance generation problem."""

from typing import Final

from moptipy.api.component import Component
from moptipy.spaces.vectorspace import VectorSpace

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.instgen.inst_decoding import InstanceDecoder
from moptipyapps.binpacking2d.instgen.instance_space import InstanceSpace


class Problem(Component):
    """An instance of the 2D Bin Packing Instance Generation Problem."""

    def __init__(self, name: str, slack: float) -> None:
        """
        Create an instance of the 2D bin packing instance generation problem.

        :param name: the name of the instance to select
        :param slack: the additional fraction of slack
        """
        super().__init__()
        #: the instance to be used as template
        self.solution_space: Final[InstanceSpace] = InstanceSpace(
            Instance.from_resource(name))
        #: the internal decoding
        self.encoding: Final[InstanceDecoder] = InstanceDecoder(
            self.solution_space)
        n: Final[int] = self.encoding.get_x_dim(slack)
        #: the search space
        self.search_space: Final[VectorSpace] = VectorSpace(n, -1.0, 1.0)

    def __str__(self) -> str:
        """
        Get the string representation of this problem.

        :return: the string representation of this problem
        """
        return (f"{self.solution_space.inst_name}_"
                f"{self.search_space.dimension}")
